$TITLE ArizonaDrugSmuggler OA4202 2010
$INLINECOM { }
 OPTIONS
   SOLPRINT =     OFF,
   DECIMALS =       1,
   LIMCOL   =      10,
   LIMROW   =      10,
   RESLIM   =      60, {max seconds}
   ITERLIM  =99999999, {max iterations}
   LP       =     CPLEX,
   MIP      =     CPLEX
 ;


$offlisting

$ontext
============================================================================
With the following few lines of code, we are reading in the network specification.
- We read in the set of nodes, from the nodes.csv file
- We read in the set of all arcs, from the arcs_set.csv file
- We define the subset of arcs where we can place roadblocks
- Finally, we read in the arc data, from the arcs_data.csv file
============================================================================
$offtext


SET
    n   nodes  /
$INCLUDE  arizona_nodes.csv
 /;

alias(n,i,j)   {(i,j) is for arc distance and flow}

SET
   arcs(i,j)         arclist  /
$ONDELIM
$INCLUDE  arizona_arcs_set.csv
$OFFDELIM
 /;

SET
   roadblockarcs(i,j)         the arcs on which we can place roadblocks  /
$ONDELIM
$INCLUDE  arizona_roadblock_arcs_set.csv
$OFFDELIM
 /;


table arcdata(i,j,*)
$ONDELIM
$INCLUDE  arizona_arcs_data.csv
$OFFDELIM
;

$ontext
============================================================================
Here, we  compute C_max and n * C_max.  The index Distance into the arcdata
table was read in by GAMS form the header of the arcs_data.csv file.
============================================================================
$offtext

SCALARS nC ;
 nC=0.0 ;

 LOOP(arcs(i,j),
   IF( nC<ABS(arcdata(i,j,'Distance')),
       nC=ABS(arcdata(i,j,'Distance')) ;
   );
 );

nC = nC * card(n);

display nC;

$ontext
============================================================================
First, lets define the typical shortest path LP.

There are three new things about our definition of the shortest path LP.
First, we have a parameter 'delay.'  This parameter stores the delay we have on
an arc, if we put up a roadblock there.  This is like the 'd_ij' we talked
about in class, except its the same for all arcs.

Let's do some estimations to see what value we should give to 'delay'.  The
distance in reality from Yuma to Gila Bend is about 117 miles.  In our data
set, the arc length between Yuma and Gila Bend is 396.772910.  That gives us
roughly 3.4 arc length units per mile.

We'll say a roadblock can delay someone for half an hour.  If traffic moves at
an average speed of 65 miles per hour, a delay of half an hour is the same as
travelling an extra 32.5 miles.  An extra 32.5 miles in our arc length units is
110.5 units.  So, we'll set the delay parameter to 110.5... that's how many
units we'll add to the length of the arc if we decide to place a roadblock
there.

Second, our definition includes the parameter xbar(i,j).  We'll use that
parameter later, to re-solve for the shortest path, given a road block plan X.
For now, we'll just set the xbar(i,j) to zero.

Finally, we'll also introduce a variable NOPATH, that makes the LP feasible
even if there is no path from the start node, Flagstaff, to the end node,
Escape.  You can think of NOPATH as adding an extra, super-expensive arc
between Flagstaff and Escape.  We'll make the cost of this arc nC, so it
doesn't get used unless there is no other path from Flagstaff to Escape.
============================================================================
$offtext

 PARAMETERS
   delay
   xbar(i,j)
 ;

 delay = 110.5;

 LOOP(arcs(i,j),
   xbar(i,j)=0 ;
 );

 VARIABLE
   Z                 objective function value
 ;
 POSITIVE VARIABLES
   Y(i,j)            flow on arc i j
   NOPATH            artificial arc from Flagstaff to Escape
 ;

 EQUATIONS
   OBJECTIVE
   NETFLOW(n)
 ;

 OBJECTIVE..
   Z =e= SUM(arcs(i,j),(  arcdata(i,j,'Distance') + delay*xbar(i,j)$(roadblockarcs(i,j) ) ) *Y(i,j) )
       + nC*NOPATH
 ;

 NETFLOW(j)..
   SUM(arcs(i,j),Y(i,j)) - SUM(arcs(j,i),Y(j,i))      {+ flow in - flow out}
     =l= +(NOPATH-1)       $diag(j,'Flagstaff')   {start node}
         +(1-NOPATH)       $diag(j,'Escape')   {end node}
 ;

 MODEL ShortestPath
 /
   OBJECTIVE
   NETFLOW
 /;

$ontext
============================================================================
Now, lets solve for the drug dealer's shortest path, when there are no
roadblocks.
============================================================================
$offtext

 SOLVE ShortestPath USING LP MINIMIZING Z;       { find shortest escape route with no roadblocks }

$ontext
============================================================================
Lets print the no roadblocks shortest path solution to a file called spstop.out
============================================================================
$offtext


 file out /spstop.out/;
 put out;

 put 'roadblock delay = ', delay,'    nC = ', nC /;
 put '****' / ;
 put 'solve ShortestPath with no roadblocks' / ;
 put 'drug smuggler responds with this shortest path:' / ;
 LOOP(arcs(i,j)$(Y.l(i,j)>0),
   put '   transit arc ',i.tl,' -> ', j.tl / ;
 );
 put 'transit cost=',Z.l:10:1 / ;


$ontext
============================================================================
Now, lets define the MIP for placing roadblocks.

We'll introduce a scalar, 'roadblocks', that specifies the number of roadblocks
we are allowed to place.

We want to maximize the drug smuggler's shortest path.  So, the problem we want
to solve is a 'max min' problem.  Like we did in class, to be able to put this
into the solver, we'll just take the dual of the inner problem, so the problem
becomes a 'max max'.  That is why in the following code, you'll see the dual of
the shortest path problem.

The binary variables X(i,j), will specify whether we place a roadblock on
arc i j.
============================================================================
$offtext


SCALARS roadblocks;

roadblocks = 0;

 POSITIVE VARIABLES
   W(n)              dual variables giving distance to node n
 ;

 BINARY VARIABLES
   X(i,j)            should we place a road block on arc i j
 ;

 EQUATIONS
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS(i,j)
   EXTRA_NOPATH_ARC_CONSTRAINT
   ROADBLOCK_LIMIT
 ;

 DUAL_OBJECTIVE..
   Z =e= W('Escape') - W('Flagstaff')   {end node minus start node}
 ;

 DUAL_CONSTRAINTS(arcs(i,j))..
   W(j)-W(i) =l= arcdata(i,j,'Distance')+delay*X(i,j)$(roadblockarcs(i,j))
 ;

 EXTRA_NOPATH_ARC_CONSTRAINT..
   W('Escape') - W('Flagstaff') =l= nC   {Comes from the extra arc we added}
 ;

 ROADBLOCK_LIMIT..
   SUM(roadblockarcs(i,j),X(i,j)) =l= roadblocks
 ;
 
 MODEL BlockPlacementMIP
 /
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS
   EXTRA_NOPATH_ARC_CONSTRAINT
   ROADBLOCK_LIMIT
 /;

$ontext
============================================================================
Now, lets solve for the best placement of road blocks.
============================================================================
$offtext

 BlockPlacementMIP.optfile = 1;

 SOLVE BlockPlacementMIP USING MIP MAXIMIZING Z;       { Find the best placement of road blocks }

$ontext
============================================================================
Lets print the roadblock placement into the output file we had opened,
spstop.out
============================================================================
$offtext

display X.l;

 put '****' / ;
 put 'roadblock plan with ', roadblocks, ' roadblocks:' / ;
 LOOP(arcs(i,j)$(X.l(i,j)>0),
   put '   roadblock on arc: ',i.tl,' -> ', j.tl / ;
 );
 put 'transit cost with roadblocks =',Z.l:10:1 / ;

$ontext
============================================================================
Finally, now that we have computed roadblock locations, lets re-solve for the
drugdealer's new best route.
============================================================================
$offtext

 LOOP(arcs(i,j),                        { transfer the roadblock solution to constant parameter }
   xbar(i,j)=X.l(i,j) ;
 );

 SOLVE ShortestPath USING LP MINIMIZING Z;       { find shortest escape route the roadblocks in place }

 put '****' / ;
 put 'solve ShortestPath with roadblocks in place' / ;
 put 'drug smuggler responds with this shortest path:' / ;
 LOOP(arcs(i,j)$(Y.l(i,j)>0),
   put '   transit arc ',i.tl,' -> ', j.tl / ;
 );
 IF(NOPATH.l>0,
   put ' There is no escape path.' / ;
 );
 put 'transit cost=',Z.l:10:1 / ;
